/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "zconf.tab.h"
#include "zconf.h"

static void conf(struct menu *menu);

static int indent = 1;
static int askold = 1;
static int conf_cnt;

static void printc(int ch)
{
	static int sep = 0;

	if (!sep) {
		putchar('[');
		sep = 1;
	} else if (ch)
		putchar('/');
	if (!ch) {
		putchar(']');
		putchar(' ');
		sep = 0;
	} else
		putchar(ch);
}

static void printo(char *o)
{
	static int sep = 0;

	if (!sep) {
		putchar('(');
		sep = 1;
	} else if (o) {
		putchar(',');
		putchar(' ');
	}
	if (!o) {
		putchar(')');
		putchar(' ');
		sep = 0;
	} else
		printf("%s", o);
}

static void readnl(void)
{
	char ch;

	do {
		if (scanf("%c", &ch) == EOF)
			exit(0);
	} while (ch != '\n');
}

int conf_string(struct menu *menu)
{
	struct symbol *sym = menu->sym;
	char line[128];

	while (1) {
		printf("%*s%s ", indent - 1, "", menu->prompt->prompt);
		printf("(%s) ", sym->name);
		if (sym->value->name)
			printf("[%s] ", sym->value->name);
		if (!sym->def_value) {
			printf("(NEW) ");
		} else {
			if (!askold) {
				printf("%s\n", sym->def_value->name);
				return 0;
			}
		}
		fgets(line, 128, stdin);
		if (line[0] == '?' && !line[1]) {
			if (menu->sym->help)
				printf("\n%s", menu->sym->help);
			continue;
		}
		if (line[0] != '\n') {
			line[strlen(line)-1] = 0;
			free(menu->prompt->def->name);
			menu->prompt->def->name = strdup(line);
		} else
			sym->def_value = menu->prompt->def;
		return 0;
	}
}

static int conf_sym(struct menu *menu)
{
	struct symbol *sym = menu->sym;
	int type, ch;

	while (1) {
		type = sym->type;
		printf("%*s%s ", indent - 1, "", menu->prompt->prompt);
		if (sym->name)
			printf("(%s) ", sym->name);
		if ((sym->flags & SYMBOL_CHOICE) && !menu->stmt)
			type = SYMBOL_BOOLEAN;
		if (type == SYMBOL_TRISTATE && calc_sym_value(modules_sym) == &symbol_no)
			type = SYMBOL_BOOLEAN;
		switch (type) {
		case SYMBOL_BOOLEAN:
			if (!(sym->flags & SYMBOL_OPTIONAL)) {
				printf("\n");
				return 0;
			}
			if (sym->value == &symbol_no) {
				printc('N');
				printc('y');
			} else {
				printc('Y');
				printc('n');
			}
			if (sym->help)
				printc('?');
			printc(0);
			break;
		case SYMBOL_TRISTATE:
			if (sym->value == &symbol_no && sym->flags & SYMBOL_OPTIONAL)
				printc('N');
			if (sym->value == &symbol_mod) {
				if (sym->max_value != &symbol_yes && !(sym->flags & SYMBOL_OPTIONAL)) {
					printf("\n");
					return 0;
				}
				printc('M');
			}
			if (sym->value == &symbol_yes && sym->max_value == &symbol_yes)
				printc('Y');

			if (sym->value != &symbol_no && sym->flags & SYMBOL_OPTIONAL)
				printc('n');
			if (sym->value != &symbol_mod)
				printc('m');
			if (sym->value != &symbol_yes && sym->max_value == &symbol_yes)
				printc('y');

			if (sym->help)
				printc('?');
			printc(0);
			break;
		}
		if (!sym->def_value)
			printf("(NEW) ");
		else if (!askold) {
			printf("%s\n", sym->value->name);
			return 0;
		}
		do {
			ch = getchar();
			if (ch == EOF)
				exit(0);
		} while (ch == ' ');
		if (ch == '\n') {
			sym->def_value = sym->value;
			return 0;
		}
		readnl();
		if (ch == '?') {
			if (sym->help)
				printf("\n%s", sym->help);
			continue;
		}
		if ((sym->flags & SYMBOL_CHOICE) && !menu->stmt) {
			struct statement *stmt;
			for (stmt = sym->stmt; stmt->next; stmt = stmt->next)
				;
			if (ch == 'n' && stmt->sym->value == &symbol_mod) {
				sym->def_value = &symbol_no;
				return 0;
			}
			if (ch == 'y') {
				stmt->def = sym;
				sym->def_value = stmt->sym->value;
				return 0;
			}
		} else switch (ch) {
		case 'n':
			if (sym->flags & SYMBOL_OPTIONAL) {
				sym->def_value = &symbol_no;
				return 0;
			}
			break;
		case 'm':
			if (type == SYMBOL_TRISTATE) {
				sym->def_value = &symbol_mod;
				return 0;
			}
			break;
		case 'y':
			if (type == SYMBOL_BOOLEAN || sym->max_value == &symbol_yes) {
				sym->def_value = &symbol_yes;
				return 0;
			}
			break;
		default:
			break;
		}
	}
}

static int conf_choice(struct menu *menu)
{
	struct symbol *sym = menu->sym;
	struct symbol *def_value = sym->def_value;
	struct menu *cmenu, *def_menu;
	int type;
	char line[128];

	type = sym->type;
	if (type == SYMBOL_TRISTATE && calc_sym_value(modules_sym) == &symbol_no)
		type = SYMBOL_BOOLEAN;
	if (sym->flags & SYMBOL_OPTIONAL || (type != SYMBOL_BOOLEAN && sym->max_value == &symbol_yes)) {
		conf_sym(menu);
		clear_sym_deps();
		calc_sym_value(menu->sym);
		if (sym->value == &symbol_no)
			return 1;
		if (sym->value == &symbol_mod)
			return 0;
	} else {
		sym->def_value = sym->value;
		if (sym->value == &symbol_mod) {
			printf("%*s%s\n", indent - 1, "", menu->prompt->prompt);
			return 0;
		}
	}
	while (1) {
		printf("%*s%s ", indent - 1, "", menu->prompt->prompt);
		def_menu = NULL;
		for (cmenu = menu->list; cmenu; cmenu = cmenu->next) {
			if (!cmenu->sym || !cmenu->prompt)
				continue;
			calc_sym_value(cmenu->sym);
			if (cmenu->sym->max_value == &symbol_no)
				continue;
			printo(cmenu->prompt->prompt);
			if (def_menu)
				continue;
			if (menu->stmt->def == cmenu->sym)
				def_menu = cmenu;
		}
		printo(NULL);
		if (def_menu)
			printf("[%s] ", def_menu->prompt->prompt);
		if (!def_value) {
			printf("(NEW) ");
		} else if (!askold && def_menu) {
			printf("%s\n", def_menu->prompt->prompt);
			return 1;
		}
		fgets(line, 128, stdin);
		if (line[0] == '?' && !line[1]) {
			if (menu->sym->help)
				printf("\n%s", menu->sym->help);
			continue;
		}
		if (line[0] != '\n') {
			line[strlen(line)-1] = 0;

			for (cmenu = menu->list; cmenu; cmenu = cmenu->next) {
				if (!cmenu->sym || !cmenu->prompt)
					continue;
				if (!strncmp(line, cmenu->prompt->prompt, strlen(line))) {
					menu->stmt->def = cmenu->sym;
					break;
				}
			}
			if (!cmenu)
				continue;
			indent += 2;
			conf(cmenu->list);
			indent -= 2;
		} else {
			if (!def_menu)
				continue;
			indent += 2;
			conf(def_menu->list);
			indent -= 2;
		}
		return 1;
	}
}

static void conf(struct menu *menu)
{
	struct statement *stmt;
	struct symbol *sym;
	struct menu *m;

	if (!menu)
		return;
	clear_sym_deps();
	sym = menu->sym;
	if ((stmt = menu->stmt)) {
		switch (stmt->token) {
		case T_MENU:
		case T_COMMENT:
			if (calc_value(stmt->dep) == &symbol_no)
				return;
			printf("%*c\n%*c %s\n%*c\n", indent, '*', indent, '*', stmt->prompt, indent, '*');
			break;
		case '^':
			calc_sym_value(sym);
			if (sym->max_value == &symbol_no)
				return;
			conf_choice(menu);
			if (sym->value != &symbol_mod)
				return;
			break;
		}
	} else if (sym) {
		if (!menu->prompt)
			goto next;
		calc_sym_value(sym);
		if (sym->max_value == &symbol_no)
			goto next;
		switch (sym->type) {
		case SYMBOL_INT:
		case SYMBOL_HEX:
		case SYMBOL_STRING:
			conf_string(menu);
			break;
		default:
			conf_sym(menu);
			break;
		}
	}

next:
	if (menu->sym)
		indent += 2;
	for (m = menu->list; m; m = m->next)
		conf(m);
	if (menu->sym)
		indent -= 2;
}

static void check_conf(struct menu *menu)
{
	struct statement *stmt;
	struct symbol *sym;
	struct menu *m;

	if (!menu)
		return;
	clear_sym_deps();
	sym = menu->sym;
	if ((stmt = menu->stmt)) {
		switch (stmt->token) {
		case '^':
			calc_sym_value(sym);
			if (sym->max_value == &symbol_no)
				return;
			if (!sym->def_value) {
				if (!conf_cnt++)
					printf("*\n* Restart config...\n*\n");
				conf(menu->parent);
			}
			if (sym->value != &symbol_mod)
				return;
			break;
		}
	} else if (sym) {
		if (!menu->prompt)
			goto next;
		calc_sym_value(sym);
		if (sym->max_value == &symbol_no)
			goto next;
		if (!sym->def_value) {
			if (!conf_cnt++)
				printf("*\n* Restart config...\n*\n");
			conf(menu->parent);
		}
	}

next:
	for (m = menu->list; m; m = m->next)
		check_conf(m);
}

int main(int ac, char **av)
{
	int silent = 0;

	zconf_initscan("config.new");
	zconfparse();
	modules_sym = lookup_symbol("MODULES");
	readconf();
	if (ac > 1 && av[1][0] == '-') {
		switch (av[1][1]) {
		case 'o':
			askold = 0;
			break;
		case 's':
			silent = 1;
			break;
		case 'h':
		case '?':
			printf("%s -o|-s\n", av[0]);
			exit(0);
		}
	}
	if (!silent)
		conf(&rootmenu);
	askold = 0;
	do {
		conf_cnt = 0;
		check_conf(&rootmenu);
	} while (conf_cnt);
	writeconf();
	return 0;
}
