/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "expr.h"

struct file *file_list;
struct file *current_file;

struct symbol symbol_yes = {
	name: "y",
	value: &symbol_yes,
	max_value: &symbol_yes,
	flags: SYMBOL_YES|SYMBOL_DEP,
}, symbol_mod = {
	name: "m",
	value: &symbol_mod,
	max_value: &symbol_mod,
	flags: SYMBOL_MOD|SYMBOL_DEP,
}, symbol_no = {
	name: "n",
	value: &symbol_no,
	max_value: &symbol_no,
	flags: SYMBOL_NO|SYMBOL_DEP,
};

struct symbol *modules_sym;

struct expr *alloc_expr_symbol(struct symbol *sym)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = T_WORD;
	e->left.sym = sym;
	return e;
}

struct expr *alloc_expr_one(int token, struct expr *ce)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = ce;
	return e;
}

struct expr *alloc_expr_two(int token, struct expr *e1, struct expr *e2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = e1;
	e->right.expr = e2;
	return e;
}

struct expr *alloc_expr_comp(int token, struct symbol *s1, struct symbol *s2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.sym = s1;
	e->right.sym = s2;
	return e;
}

struct expr *copy_expr(struct expr *org)
{
	struct expr *e;

	if (!org)
		return NULL;

	e = malloc(sizeof(*org));
	memcpy(e, org, sizeof(*org));
	switch (org->token) {
	case T_WORD:
		e->left = org->left;
		break;
	case '!':
		e->left.expr = copy_expr(org->left.expr);
		break;
	case '=':
	case T_UNEQUAL:
		e->left.sym = org->left.sym;
		e->right.sym = org->right.sym;
		break;
	case T_OR:
	case T_AND:
	case '&':
	case '^':
	case T_IF:
	case T_ELSE:
		e->left.expr = copy_expr(org->left.expr);
		e->right.expr = copy_expr(org->right.expr);
		break;
	default:
		printf("can't copy token %d\n", e->token);
		free(e);
		e = NULL;
		break;
	}

	return e;
}

void free_expr(struct expr *e)
{
	if (!e)
		return;

	switch (e->token) {
	case T_WORD:
		break;
	case '!':
		free_expr(e->left.expr);
		return;
	case '=':
	case T_UNEQUAL:
		break;
	case T_OR:
	case T_AND:
	case '&':
	case T_IF:
	case T_ELSE:
		free_expr(e->left.expr);
		free_expr(e->right.expr);
		break;
	default:
		printf("how to free token %d?\n", e->token);
		break;
	}
	free(e);
}

struct symbol *calc_value_and(struct symbol *val1, struct symbol *val2)
{
	if (val1 == &symbol_no || val2 == &symbol_no)
		return &symbol_no;
	if (val1 == &symbol_mod || val2 == &symbol_mod)
		return &symbol_mod;
	return &symbol_yes;
}

struct symbol *calc_value_or(struct symbol *val1, struct symbol *val2)
{
	if (val1 == &symbol_yes || val2 == &symbol_yes)
		return &symbol_yes;
	if (val1 == &symbol_mod || val2 == &symbol_mod)
		return &symbol_mod;
	return &symbol_no;
}

struct symbol *calc_value(struct expr *e)
{
	struct symbol *max1, *max2;

	if (!e)
		return &symbol_yes;

	switch (e->token) {
	case T_WORD:
		if (e->left.sym->flags & SYMBOL_CONST)
			return e->left.sym;
		if (e->left.sym->type == SYMBOL_UNKNOWN)
			return &symbol_yes;
		return calc_sym_value(e->left.sym);
	case '&':
		max1 = calc_value(e->left.expr);
		max2 = calc_value(e->right.expr);
		if (max1 == &symbol_no || max2 == &symbol_no)
			return &symbol_no;
		if (max1 == &symbol_mod || max2 == &symbol_mod)
			return &symbol_mod;
		return &symbol_yes;
	case '|':
		max1 = calc_value(e->left.expr);
		max2 = calc_value(e->right.expr);
		if (max1 == &symbol_yes || max2 == &symbol_yes)
			return &symbol_yes;
		if (max1 == &symbol_mod || max2 == &symbol_mod)
			return &symbol_mod;
		return &symbol_no;
	case '!':
		max1 = calc_value(e->left.expr);
		if (max1 == &symbol_yes)
			return &symbol_no;
		if (max1 == &symbol_no)
			return &symbol_yes;
		return max1;
	case '=':
		max1 = calc_sym_value(e->left.sym);
		max2 = calc_sym_value(e->right.sym);
		if (max1 == max2 || !strcmp(max1->name, max2->name))
			return &symbol_yes;
		return &symbol_no;
	case T_UNEQUAL:
		max1 = calc_sym_value(e->left.sym);
		max2 = calc_sym_value(e->right.sym);
		if (max1 == max2 || !strcmp(max1->name, max2->name))
			return &symbol_no;
		return &symbol_yes;
	default:
		printf("calc_deps: %d?\n", e->token);
		return &symbol_no;
	}
}

struct symbol *calc_sym_value(struct symbol *sym)
{
	struct statement *stmt;
	struct symbol *max = &symbol_no;
	struct symbol *sym_max;

	if (sym->flags & SYMBOL_DEP)
		return sym->value;
	switch (sym->type) {
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			if (stmt->prompt && stmt->def) {
				sym->value = stmt->def;
				break;
			}
		}
		break;
	default:
		sym->value = &symbol_no;
		break;
	}
	sym->flags |= SYMBOL_DEP;
	sym->flags &= ~SYMBOL_WRITE;
	sym_max = calc_value(sym->dep);
	if ((sym->flags & SYMBOL_CHOICE) && sym->name) {
		for (stmt = sym->stmt; stmt->next; stmt = stmt->next) {
			if (!stmt->prompt)
				continue;
			max = calc_value_or(max, calc_value(stmt->dep));
		}
		max = calc_value_and(max, sym_max);
		sym->max_value = calc_value_and(max, calc_sym_value(stmt->sym));
		sym->value = sym->def_value ? sym->def_value : &symbol_no;
		sym->value = calc_value_and(sym->value, sym->max_value);
		if (stmt->sym->value == &symbol_yes)
			sym->value = stmt->def == sym ? &symbol_yes : &symbol_no;
		if (stmt->sym->max_value != &symbol_no)
			sym->flags |= SYMBOL_WRITE;
		return sym->value;
	}
	if (sym->name) {
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			if (!stmt->prompt)
				continue;
			max = calc_value_or(max, calc_value(stmt->dep));
		}
		max = calc_value_and(max, sym_max);
	} else
		max = sym_max;
	sym->max_value = max;
	if (max == &symbol_no) {
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			if (stmt->prompt)
				continue;
			max = calc_value_and(sym_max, calc_value(stmt->dep));
			if (max != &symbol_no) {
				sym->flags |= SYMBOL_WRITE;
				switch (sym->type) {
				case SYMBOL_BOOLEAN:
				case SYMBOL_TRISTATE:
					sym->value = calc_sym_value(stmt->def);
					break;
				case SYMBOL_INT:
				case SYMBOL_HEX:
				case SYMBOL_STRING:
					sym->value = stmt->def;
					break;
				}
				break;
			}
		}
	} else {
		sym->flags |= SYMBOL_WRITE;
		switch (sym->type) {
		case SYMBOL_INT:
		case SYMBOL_HEX:
		case SYMBOL_STRING:
			if (sym->def_value)
				sym->value = sym->def_value;
			return sym->value;
		}
		if (max == &symbol_yes) {
			if (sym->def_value)
				sym->value = sym->def_value;
		} else if (max == &symbol_mod) {
			if (sym->def_value && sym->def_value != &symbol_no)
				sym->value = &symbol_mod;
		}
		if (!(sym->flags & SYMBOL_OPTIONAL) && sym->value == &symbol_no)
			sym->value = max;
	}
	if (sym->value == &symbol_mod) {
		if (sym->type == SYMBOL_TRISTATE && calc_sym_value(modules_sym) == &symbol_no)
			sym->value = &symbol_yes;
	}
	return sym->value;
}

void clear_sym_deps(void)
{
	struct symbol *sym;
	int i;

	for_all_symbols(i, sym)
		sym->flags &= ~SYMBOL_DEP;
}

int contains_symbol(struct expr *dep, struct symbol *sym)
{
	if (!dep)
		return 0;

	switch (dep->token) {
	case T_AND:
	case T_OR:
	case '&':
	case '|':
		return contains_symbol(dep->left.expr, sym) ||
		       contains_symbol(dep->right.expr, sym);
	case T_WORD:
		return dep->left.sym == sym;
	case '=':
	case T_UNEQUAL:
		return dep->left.sym == sym;
	case '!':
		return contains_symbol(dep->left.expr, sym);
	}
	return 0;
}

static inline int token_gt(int t1, int t2)
{
#if 0
	return 1;
#else
	switch (t1) {
	case '=':
	case T_UNEQUAL:
		if (t2 == '!')
			return 1;
	case '!':
		if (t2 == T_AND || t2 == T_IF || t2 == T_ELSE || t2 == '&')
			return 1;
	case '&':
	case T_IF:
	case T_ELSE:
	case T_AND:
		if (t2 == T_OR || t2 == '|')
			return 1;
	case '|':
	case T_OR:
		if (t2 == '^')
			return 1;
	case '^':
		if (t2 == 0)
			return 1;
	case 0:
		return 0;
	}
	printf("[%dgt%d?]", t1, t2);
	return 0;
#endif
}

int print_type = 1;

static void print_sym(FILE *out, struct symbol *sym)
{
	fprintf(out, "%s", sym->name);
	if (print_type)
		fprintf(out, "%s",
			sym->type == SYMBOL_BOOLEAN ? "?" :
			sym->type == SYMBOL_TRISTATE ? "??" :
			sym->type == SYMBOL_HEX ? "h" :
			sym->type == SYMBOL_INT ? "i" :
			sym->type == SYMBOL_STRING ? "s" :
		       	"");
}

void fprint_expr(FILE *out, struct expr *e, int prevtoken)
{
	if (!e) {
		fprintf(out, "<none>");
		return;
	}

	switch (e->token) {
	case T_WORD:
		print_sym(out, e->left.sym);
		break;
	case '!':
		if (token_gt(prevtoken, '!'))
			fprintf(out, "(");
		fprintf(out, "!");
		fprint_expr(out, e->left.expr, '!');
		if (token_gt(prevtoken, '!'))
			fprintf(out, ")");
		break;
	case '=':
		if (token_gt(prevtoken, '='))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, '='))
			fprintf(out, ")");
		break;
	case T_UNEQUAL:
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "!=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, ")");
		break;
	case T_OR:
		if (token_gt(prevtoken, T_OR))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, T_OR);
		fprintf(out, " || ");
		fprint_expr(out, e->right.expr, T_OR);
		if (token_gt(prevtoken, T_OR))
			fprintf(out, ")");
		break;
	case T_AND:
		if (token_gt(prevtoken, T_AND))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, T_AND);
		fprintf(out, " && ");
		fprint_expr(out, e->right.expr, T_AND);
		if (token_gt(prevtoken, T_AND))
			fprintf(out, ")");
		break;
	case '|':
		if (token_gt(prevtoken, '|'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '|');
		fprintf(out, " || ");
		fprint_expr(out, e->right.expr, '|');
		if (token_gt(prevtoken, '|'))
			fprintf(out, ")");
		break;
	case '&':
		if (token_gt(prevtoken, '&'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '&');
		if (print_type)
			fprintf(out, " & ");
		else
			fprintf(out, " && ");
		fprint_expr(out, e->right.expr, '&');
		if (token_gt(prevtoken, '&'))
			fprintf(out, ")");
		break;
	case '^':
		if (e->left.expr) {
			fprint_expr(out, e->left.expr, '^');
			fprintf(out, " ^ ");
		}
		fprintf(out, "%s", e->right.sym->name);
		break;
	case T_IF:
		if (token_gt(prevtoken, T_IF))
			fprintf(out, "[");
		if (e->right.expr) {
			fprint_expr(out, e->right.expr, T_IF);
			fprintf(out, " && ");
		}
		fprint_expr(out, e->left.expr, T_IF);
		if (token_gt(prevtoken, T_IF))
			fprintf(out, "]");
		break;
	case T_ELSE:
		if (token_gt(prevtoken, T_ELSE))
			fprintf(out, "[");
		fprintf(out, "[");
		if (e->right.expr) {
			fprint_expr(out, e->right.expr, T_ELSE);
			fprintf(out, " && ");
		}
		fprintf(out, "!");
		fprint_expr(out, e->left.expr, '!');
		if (token_gt(prevtoken, T_ELSE))
			fprintf(out, "]");
		break;
	default:
		fprintf(out, "<unknown token %d>", e->token);
		break;
	}
}

void print_expr(int mask, struct expr *e, int prevtoken)
{
	if (!(cdebug & mask))
		return;
	fprint_expr(stdout, e, prevtoken);
}

struct file *lookup_file(char *name)
{
	struct file *file;

	for (file = file_list; file; file = file->next) {
		if (!strcmp(name, file->name))
			return file;
	}

	file = malloc(sizeof(*file));
	memset(file, 0, sizeof(*file));
	file->name = strdup(name);
	file->next = file_list;
	file_list = file;
	return file;
}
