%{
/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zconf.h"

#define printd(mask, fmt...) if (cdebug & (mask)) printf(fmt)

#define PRINTD		0x0001
#define DEBUG_PARSE	0x0002

int cdebug = PRINTD;

extern int zconflex(void);
static void zconferror(const char *err);

struct menu rootmenu;
static struct menu *current_menu = &rootmenu;

struct symbol *symbol_hash[257];
static struct symbol *current_symbol;
static struct symbol *current_choice;
static struct statement *current_stmt;

void reindent_menu(void);
struct statement *create_stmt(int token);
struct statement *add_stmt(char *prompt, struct symbol *def, struct expr *dep);
struct menu *add_menu(struct symbol *sym, struct statement *stmt);

#define YYERROR_VERBOSE
%}

%union
{
	char *string;
	struct symbol *symbol;
	struct expr *expr;
	struct statement *stmt;
}

%token T_MAINMENU
%token T_MENU
%token T_ENDMENU
%token T_SOURCE
%token T_CHOICE
%token T_ENDCHOICE
%token T_COMMENT
%token T_CONFIG
%token T_HELP
%token <string> T_HELPTEXT
%token T_IF 304
%token T_DEPENDS
%token T_OPTIONAL
%token T_DEFAULT
%token T_PROMPT
%token T_TRISTATE, T_BOOLEAN, T_INT, T_HEX, T_STRING
%token <string> T_WORD 300, T_STRING
%token T_UNEQUAL 301

%left '|'
%left '&'
%left '=' T_UNEQUAL
%nonassoc '!'

%type <string> prompt
%type <symbol> symbol, symbol_data
%type <expr> expr, if_expr
%type <stmt> statement, menu_stmt, choice_stmt

%%
input:  block
;

block:	  /* empty */
	| block menu_stmt
	| block choice_stmt
	| block config_stmt
	| block statement
;

menu_stmt: T_MENU prompt '\n'
{
	$$ = current_stmt = create_stmt(T_MENU);
	$$->prompt = $2;
	add_menu(NULL, $$);
	printd(DEBUG_PARSE, ":%d:menu\n", lineno());
}
	stmt_dep_opt
{
	reindent_menu();
}
	menu_block T_ENDMENU '\n'
{
	while (!current_menu->stmt || current_menu->stmt->token != T_MENU)
		current_menu = current_menu->parent;
	current_menu = current_menu->parent;
	printd(DEBUG_PARSE, ":%d:endmenu\n", lineno());
};

menu_block:	  /* empty */
		| menu_block menu_stmt
		| menu_block choice_stmt
		| menu_block config_stmt
		| menu_block statement
;

choice_stmt: T_CHOICE '\n'
{
	$$ = current_stmt = create_stmt('^');
	current_choice = alloc_symbol(NULL);
	current_choice->type = SYMBOL_BOOLEAN;
	current_choice->stmt = $$;
	current_choice->flags = SYMBOL_CHOICE;
	current_choice->next = symbol_hash[256];
	symbol_hash[256] = current_choice;
	$$->sym = current_choice;
	add_menu(current_choice, $$);
	printd(DEBUG_PARSE, ":%d:choice\n", lineno());
}
	choice_args
{
	reindent_menu();
}
	choice_block T_ENDCHOICE '\n'
{
	struct menu *menu;
	struct symbol *def;

	while (current_menu->sym != current_choice)
		current_menu = current_menu->parent;

	//if (!(current_choice->flags & SYMBOL_OPTIONAL))
	//	if (!current_choice->def_value || current_choice->def_value == &symbol_no)
	//		current_choice->def_value = &symbol_yes;
	def = current_menu->stmt->def;
	if (def) for (menu = current_menu->list; menu; menu = menu->next) {
		if (!menu->sym || !menu->prompt)
			continue;
		//if (!def) {
		//	current_menu->stmt->def = menu->sym;
		//	break;
		//}
		if (!strncmp(def->name, menu->prompt->prompt, strlen(def->name))) {
			current_menu->stmt->def = menu->sym;
			break;
		}
	}

	current_menu = current_menu->parent;
	current_choice = NULL;
	printd(DEBUG_PARSE, ":%d:endchoice\n", lineno());
};

choice_args: /* empty */
;

choice_args: choice_args T_DEPENDS expr '\n'
{
	current_choice->dep = !current_choice->dep ? $3 :
		alloc_expr_two('&', current_choice->dep, $3);
	printd(DEBUG_PARSE, ":%d:depends\n", lineno());
};

choice_args: choice_args T_PROMPT prompt '\n'
{
	current_choice->stmt->prompt = $3;
	current_menu->prompt = current_choice->stmt;
	printd(DEBUG_PARSE, ":%d:prompt\n", lineno());
};

choice_args: choice_args T_OPTIONAL '\n'
{
	current_choice->flags |= SYMBOL_OPTIONAL;
	current_choice->stmt->dep = alloc_expr_one('^', current_choice->stmt->dep);
	current_choice->stmt->dep->right.sym = &symbol_no;
	printd(DEBUG_PARSE, ":%d:optional\n", lineno());
};

choice_args: choice_args T_DEFAULT symbol '\n'
{
	current_choice->stmt->def = $3;
	printd(DEBUG_PARSE, ":%d:default\n", lineno());
};

choice_args: choice_args T_HELP '\n'
{
	printd(DEBUG_PARSE, ":%d:help\n", lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_choice->help = $5;
};

choice_block:	  /* empty */
		| choice_block config_stmt
		| choice_block '\n'
;

statement: '\n'
{
	$$ = NULL;
};

statement: T_MAINMENU prompt '\n'
{
	$$ = create_stmt(T_MAINMENU);
	$$->prompt = $2;
	printd(DEBUG_PARSE, ":%d:menu\n", lineno());
};

statement: T_SOURCE prompt '\n'
{
	$$ = create_stmt(T_SOURCE);
	$$->prompt = $2;
	printd(DEBUG_PARSE, ":%d:source %s\n", lineno(), $2);
};

statement: T_COMMENT prompt '\n'
{
	$$ = current_stmt = create_stmt(T_COMMENT);
	$$->prompt = $2;
	add_menu(NULL, $$);
	printd(DEBUG_PARSE, ":%d:comment\n", lineno());
}
	stmt_dep_opt
{
	reindent_menu();
};

stmt_dep_opt:	  /* empty */
		| T_DEPENDS expr '\n'
{
	current_stmt->dep = !current_stmt->dep ? $2 :
		alloc_expr_two('&', current_stmt->dep, $2);
	printd(DEBUG_PARSE, ":%d:depends\n", lineno());
};

config_stmt: T_CONFIG T_WORD '\n'
{
	printd(DEBUG_PARSE, ":%d:config %s\n", lineno(), $2);
	current_symbol = lookup_symbol($2);
	current_symbol->menu = current_menu;
	current_symbol->flags |= SYMBOL_OPTIONAL;
	add_menu(current_symbol, NULL);
}
	config_args
{
	//if (!current_symbol->stmt)
	//	add_stmt(NULL, NULL, NULL);
	if (current_menu->prompt && !current_menu->prompt->def) {
		switch (current_symbol->type) {
		case SYMBOL_INT:
		case SYMBOL_HEX:
		case SYMBOL_STRING:
			current_menu->prompt->def = alloc_symbol(NULL);
		}
	}
	reindent_menu();
	if (current_menu->parent->stmt && current_menu->parent->stmt->token == '^') {
		struct expr **ep;
		struct statement **sp;

		current_symbol->flags |= SYMBOL_CHOICE;
		for (sp = &current_symbol->stmt; *sp; sp = &(*sp)->next)
			;
		*sp = current_choice->stmt;
		for (ep = &current_choice->stmt->dep; *ep; ep = &(*ep)->left.expr)
			;
		*ep = alloc_expr_one('^', NULL);
		(*ep)->right.sym = current_symbol;
		if (current_symbol->type == SYMBOL_TRISTATE)
			current_choice->type = SYMBOL_TRISTATE;
	}
	current_symbol = NULL;
};

config_args: /* empty */
;

config_args: config_args T_DEPENDS expr '\n'
{
	struct symbol *sym = current_symbol;
	sym->dep = sym->dep ? alloc_expr_two('&', sym->dep, $3) : $3;
	printd(DEBUG_PARSE, ":%d:depends\n", lineno());
};

config_args: config_args T_TRISTATE '\n'
{
	current_symbol->type = SYMBOL_TRISTATE;
	printd(DEBUG_PARSE, ":%d:tristate\n", lineno());
};

config_args: config_args T_TRISTATE prompt if_expr '\n'
{
	current_symbol->type = SYMBOL_TRISTATE;
	current_menu->prompt = add_stmt($3, NULL, $4);
	printd(DEBUG_PARSE, ":%d:tristate\n", lineno());
};

config_args: config_args T_BOOLEAN '\n'
{
	current_symbol->type = SYMBOL_BOOLEAN;
	printd(DEBUG_PARSE, ":%d:boolean\n", lineno());
};

config_args: config_args T_BOOLEAN prompt if_expr '\n'
{
	current_symbol->type = SYMBOL_BOOLEAN;
	current_menu->prompt = add_stmt($3, NULL, $4);
	printd(DEBUG_PARSE, ":%d:boolean\n", lineno());
};

config_args: config_args T_INT '\n'
{
	current_symbol->type = SYMBOL_INT;
	printd(DEBUG_PARSE, ":%d:int\n", lineno());
};

config_args: config_args T_INT prompt symbol_data if_expr '\n'
{
	current_symbol->type = SYMBOL_INT;
	current_menu->prompt = add_stmt($3, $4, $5);
	printd(DEBUG_PARSE, ":%d:int\n", lineno());
};

config_args: config_args T_HEX '\n'
{
	current_symbol->type = SYMBOL_HEX;
	printd(DEBUG_PARSE, ":%d:hex\n", lineno());
};

config_args: config_args T_HEX prompt symbol_data if_expr '\n'
{
	current_symbol->type = SYMBOL_HEX;
	current_menu->prompt = add_stmt($3, $4, $5);
	printd(DEBUG_PARSE, ":%d:hex\n", lineno());
};

config_args: config_args T_STRING '\n'
{
	current_symbol->type = SYMBOL_STRING;
	printd(DEBUG_PARSE, ":%d:string\n", lineno());
};

config_args: config_args T_STRING prompt symbol_data if_expr '\n'
{
	current_symbol->type = SYMBOL_STRING;
	current_menu->prompt = add_stmt($3, $4, $5);
	printd(DEBUG_PARSE, ":%d:string\n", lineno());
};

config_args: config_args T_PROMPT prompt symbol_data if_expr '\n'
{
	current_menu->prompt = add_stmt($3, $4, $5);
	printd(DEBUG_PARSE, ":%d:prompt\n", lineno());
};

config_args: config_args T_DEFAULT symbol if_expr '\n'
{
	add_stmt(NULL, $3, $4);
	printd(DEBUG_PARSE, ":%d:default\n", lineno());
};

config_args: config_args T_HELP '\n'
{
	printd(DEBUG_PARSE, ":%d:help\n", lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_symbol->help = $5;
};

prompt:	  T_WORD
	| T_STRING
;

if_expr:  /* empty */			{ $$ = NULL; }
	| T_IF expr			{ $$ = $2; }
;

expr:	  symbol			{ $$ = alloc_expr_symbol($1); }
	| symbol '=' symbol		{ $$ = alloc_expr_comp('=', $1, $3); }
	| symbol T_UNEQUAL symbol	{ $$ = alloc_expr_comp(T_UNEQUAL, $1, $3); }
	| expr '|' expr			{ $$ = alloc_expr_two('|', $1, $3); }
	| expr '&' expr			{ $$ = alloc_expr_two('&', $1, $3); }
	| '(' expr ')'			{ $$ = $2; }
	| '!' expr			{ $$ = alloc_expr_one('!', $2); }
;

symbol:	  T_WORD	{ $$ = lookup_symbol($1); free($1); }
	| T_STRING	{ $$ = lookup_symbol($1); free($1); }
;

symbol_data:	  /* empty */		{ $$ = NULL; }
		| T_WORD		{ $$ = alloc_symbol($1); }
		| T_STRING		{ $$ = alloc_symbol($1); }
;
%%
static void zconferror(const char *err)
{
	printf(":%d:%s\n", lineno(), err);
	exit(1);
}

struct symbol *alloc_symbol(char *name)
{
	struct symbol *symbol;

	symbol = malloc(sizeof(*symbol));
	memset(symbol, 0, sizeof(*symbol));
	symbol->name = name;
	symbol->type = SYMBOL_UNKNOWN;

	return symbol;
}

struct symbol *lookup_symbol(char *name)
{
	struct symbol *symbol;
	char *ptr;
	int hash = 0;

	//printf("lookup: %s -> ", name);
	if (name[0] && !name[1]) {
		switch (name[0]) {
		case 'y': return &symbol_yes;
		case 'm': return &symbol_mod;
		case 'n': return &symbol_no;
		}
	}
	for (ptr = name; *ptr; ptr++)
		hash += *ptr;

	for (symbol = symbol_hash[hash&255]; symbol; symbol = symbol->next) {
		if (!strcmp(symbol->name, name)) {
			//printf("h:%p\n", symbol);
			return symbol;
		}
	}

	symbol = malloc(sizeof(*symbol));
	memset(symbol, 0, sizeof(*symbol));
	symbol->name = strdup(name);
	symbol->type = SYMBOL_UNKNOWN;

	symbol->next = symbol_hash[hash&255];
	symbol_hash[hash&255] = symbol;

	//printf("n:%p\n", symbol);
	return symbol;
}

void reindent_menu(void)
{
	struct menu *parent, **menup;
	struct expr *dep;

	if (current_menu->sym)
		dep = current_menu->sym->dep;
	else
		dep = current_menu->stmt->dep;
	parent = current_menu->parent;
	while (parent->parent) {
		if (parent->stmt && (parent->stmt->token == T_MENU || parent->stmt->token == '^'))
			break;
		if (parent->prompt && parent->sym && contains_symbol(dep, parent->sym))
			break;
		parent = parent->parent;
	}
	current_menu->parent = parent;
	for (menup = &parent->list; *menup; menup = &(*menup)->next)
		;
	*menup = current_menu;
}

struct statement *create_stmt(int token)
{
	struct statement *stmt;

	stmt = malloc(sizeof(*stmt));
	memset(stmt, 0, sizeof(*stmt));
	stmt->token = token;
	stmt->file = current_file;
	stmt->lineno = lineno();
	if (!current_file->stmt)
		current_file->stmt = stmt;
	if (current_file->last_stmt)
		current_file->last_stmt->next_pos = stmt;
	current_file->last_stmt = stmt;

	return stmt;
}

struct statement *add_stmt(char *prompt, struct symbol *def, struct expr *dep)
{
	struct symbol *sym = current_symbol;
	struct statement *stmt = create_stmt(0);
	struct statement **stmtp;

	stmt->prompt = prompt;
	stmt->def = def;
	stmt->dep = dep;
	stmt->sym = sym;

	stmtp = &sym->stmt;
	while (*stmtp && (*stmtp)->token != '^')
		stmtp = &(*stmtp)->next;
	stmt->next = (*stmtp);
	(*stmtp) = stmt;

	return stmt;
}

struct menu *add_menu(struct symbol *sym, struct statement *stmt)
{
	struct menu *menu;

	menu = malloc(sizeof(*menu));
	memset(menu, 0, sizeof(*menu));
	menu->sym = sym;
	menu->stmt = stmt;

	menu->parent = current_menu;
	current_menu = menu;

	return menu;
}

void print_quoted_string(FILE *out, const char *str)
{
	const char *p;
	int len;

	putc('"', out);
	while ((p = strchr(str, '"'))) {
		len = p - str;
		if (len)
			fprintf(out, "%.*s", len, str);
		fputs("\\\"", out);
		str += len + 1;
	}
	fputs(str, out);
	putc('"', out);
}

void do_print_symbol(FILE *out, struct symbol *sym)
{
	struct statement *stmt;

	sym->flags |= SYMBOL_PRINTED;

	fprintf(out, "config %s\n", sym->name);
	if (sym->type == SYMBOL_BOOLEAN)
		fputs("  boolean\n", out);
	else if (sym->type == SYMBOL_TRISTATE)
		fputs("  tristate\n", out);
	if (!expr_is_yes(sym->dep)) {
		fputs("  depends ", out);
		fprint_expr(out, sym->dep, 0);
		fputc('\n', out);
	}
	for (stmt = sym->stmt; stmt; stmt = stmt->next) {
		if (stmt->token == '^')
			continue;
		if (stmt->prompt) {
			char *p = stmt->prompt;
			while (*p == ' ')
				p++;
			fputs("  prompt ", out);
			print_quoted_string(out, p);
			if (stmt->def) {
				fputc(' ', out);
				print_quoted_string(out, stmt->def->name);
			}
			if (!expr_is_yes(stmt->dep)) {
				fputs(" if ", out);
				fprint_expr(out, stmt->dep, 0);
			}
			fputc('\n', out);
		} else if (stmt->def) {
			if (sym->flags & SYMBOL_CHOICE)
				fputs( "  prompt ", out);
			else
				fputs( "  default ", out);
			print_quoted_string(out, stmt->def->name);
			if (!expr_is_yes(stmt->dep)) {
				fputs(" if ", out);
				fprint_expr(out, stmt->dep, 0);
			}
			fputc('\n', out);
		} else if (!expr_is_yes(stmt->dep)) {
			fputs("  depends ", out);
			fprint_expr(out, stmt->dep, 0);
			fputc('\n', out);
		}
	}
	if (sym->help) {
		int len = strlen(sym->help);
		while (sym->help[--len] == '\n')
			sym->help[len] = 0;
		fprintf(out, "  help\n%s\n", sym->help);
	}
	fputc('\n', out);
}

void print_choice(FILE *out, struct symbol *sym)
{
	struct statement *stmt;
	struct expr *lst;

	for (stmt = sym->stmt; stmt; stmt = stmt->next)
		if (stmt->token == '^')
			break;
	if (!stmt)
		//warn?
		return;

	fputs("choice\n", out);
	if (stmt->prompt) {
		char *p = stmt->prompt;
		while (*p == ' ')
			p++;
		fputs("  prompt ", out);
		print_quoted_string(out, p);
		fputc('\n', out);
	}

	for (lst = stmt->dep ;lst; lst = lst->left.expr) {
		sym = lst->right.sym;
		if (sym == &symbol_no) {
			fputs("  optional\n", out);
			continue;
		}
	}
	if (!expr_is_yes(stmt->dep)) {
		fputs("  depends ", out);
		fprint_expr(out, stmt->dep, 0);
		fputc('\n', out);
	}
	if (stmt->def) {
		fputs("  default ", out);
		print_quoted_string(out, stmt->def->name);
		fputc('\n', out);
	}

	fputc('\n', out);

	for (lst = stmt->dep; lst; lst = lst->left.expr) {
		sym = lst->right.sym;
		if (sym != &symbol_no)
			do_print_symbol(out, sym);
	}

	fputs("endchoice\n\n", out);
}

void print_symbols(FILE *out, struct symbol *sym)
{
	if (sym->flags & SYMBOL_PRINTED)
		return;
	sym->flags |= SYMBOL_PRINTED;

	if (sym->flags & SYMBOL_CHOICE) {
		print_choice(out, sym);
		return;
	}
	do_print_symbol(out, sym);
}

void zconfdump(FILE *out)
{
	//struct file *file;
	struct statement *stmt;
	struct symbol *sym;
	struct menu *menu;

	print_type = 0;

#if 0
	for (file = file_list; file; file = file->next) {
		fprintf(out, "\nfile %s\n\n", file->name);
		for (stmt = file->stmt; stmt; stmt = stmt->next_pos) {
			sym = stmt->sym;
			if (!sym) {
				switch (stmt->token) {
				case T_MAINMENU:
					fputs("\nmainmenu ", out);
					print_quoted_string(out, stmt->prompt);
					fputs("\n\n", out);
					break;
				case T_COMMENT:
					fputs("\ncomment ", out);
					print_quoted_string(out, stmt->prompt);
					fputs("\n\n", out);
					break;
				case T_MENU:
					fputs("\nmenu ", out);
					print_quoted_string(out, stmt->prompt);
					fputs("\n\n", out);
					break;
				case T_ENDMENU:
					fputs("\nendmenu\n\n", out);
					break;
				case T_SOURCE:
					fputs("\nsource ", out);
					print_quoted_string(out, stmt->prompt);
					fputs("\n\n", out);
					break;
				}
			} else
				print_symbols(out, sym);
		}
	}
#else
	menu = rootmenu.list;
	while (menu) {
		if ((stmt = menu->stmt)) {
			switch (stmt->token) {
			case T_MAINMENU:
				fputs("\nmainmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n\n", out);
				break;
			case T_COMMENT:
				fputs("\ncomment ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n\n", out);
				break;
			case T_MENU:
				fputs("\nmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n\n", out);
				break;
			case T_ENDMENU:
				fputs("\nendmenu\n\n", out);
				break;
			case T_SOURCE:
				fputs("\nsource ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n\n", out);
				break;
			}
		} else if ((sym = menu->sym))
			print_symbols(out, menu->sym);

		if (menu->list)
			menu = menu->list;
		else if (menu->next)
			menu = menu->next;
		else while ((menu = menu->parent)) {
			if (menu->next) {
				menu = menu->next;
				break;
			}
		}
	}
#endif
}
